from .rule_algorithm import RuleBuilderAlgorithm
from .classifier import Classifier
from ..data_structures import ClassAssocationRule, Antecedent, Consequent

import random

import collections

class M2Algorithm(RuleBuilderAlgorithm):
    """
    Implementation of M2 Algorithm for CBA.
    """
    
    def build(self):

        self.rules.sort(reverse=True)
        
        self.dataset_frozen = self.dataset
        self.dataset_len = len(self.dataset_frozen)

        # set of crules that have higher precedence
        # that their corresponding wrules
        self.Q = set()
        
        # set of all crules
        self.U = set()
        
        # set of conflicting rules
        self.A = set()
        
        self.classifier = []
        
        self.stage1()
        self.stage2()
        self.stage3()
        
        clf = Classifier()
        clf.rules = self.classifier
        clf.default_class = self.default_class
        self.calculate_default_class_properties(clf)
        
        return clf
    
        
    def stage1(self):
        for datacase in self.dataset_frozen:
            # finds the highest precedence crules and wrules
            crule, wrule = self.maxcoverrule(datacase, self.rules)
        
            if crule is None:
                crule = self.emptyrule()
                
            if wrule is None:
                wrule = self.emptyrule()
                
            self.U.add(crule)
            
            crule.class_cases_covered.update([datacase.class_val.value])
            
            if crule > wrule:
                self.Q.add(crule)
                crule.marked = True
            else:
                structure = (datacase, datacase.class_val.value, crule, wrule)
                self.A.add(structure)
                
            
                
    
    def stage2(self):
        
        for conflicting_struct in self.A:
            datacase, clazz, crule, wrule = conflicting_struct
            
            
            if wrule.marked:
                crule.class_cases_covered[clazz] -= 1
                wrule.class_cases_covered[clazz] += 1
            
            else:
                wset = self.allcover_rules(self.U, datacase, crule)
                for w in wset:
                    w.replace.add((crule, datacase, clazz))
                    w.class_cases_covered[clazz] += 1
                    
                self.Q = self.Q.union(wset)
        
        
    def stage3(self):
        Qlist = sorted(self.Q, reverse=True)

        rule_errors = 0
        rule_supcount = 0
        total_errors_list = []
        default_classes_list = []
        rules_list = []
        
        # class distribution
        classdist = collections.Counter(map(lambda d: d.class_val.value, self.dataset_frozen))
        classdist_keys = list(classdist.keys())
        
        for rule in Qlist:
            if rule.class_cases_covered[rule.consequent.value] > 0:
                for (rule_replace, dcase, clazz) in rule.replace:
                    if dcase.alreadycovered == True:
                        rule.class_cases_covered[clazz] -= 1
                    else:
                        dcase.alreadycovered = True
                        rule_replace.class_cases_covered[clazz] -= 1
                
                rule_errors += self.errors_of_rule(rule)
                rule_supcount += rule.support_count
                
                classdist = self.update_class_distr(classdist, rule)
                
                default_class = self.select_default_class(classdist)
                default_class_count = default_class[1]
                default_class_label = default_class[0]
                
                default_errors = self.dataset_len - rule_supcount - default_class_count
                
                total_errors = rule_errors + default_errors
                
                rules_list.append(rule)
                default_classes_list.append(default_class_label)
                total_errors_list.append(total_errors)
                
        
        if len(total_errors_list) != 0:
            min_value = min(total_errors_list)
            
            min_indices = [ idx for (idx, err_num) in enumerate(total_errors_list) if err_num == min_value ]
            min_idx = min_indices[0]
            
            final_classifier = [ rule for rule in rules_list[:min_idx + 1] ]
            default_class = default_classes_list[min_idx]

            if not default_class:
                i = min_idx
                while not default_class:
                    i -= 1
                    default_class = default_classes_list[i]

            self.classifier = final_classifier
            self.default_class = default_class
            self.default_class_attribute = classdist_keys[0][0]
        else:
            possible_default_classes = list(classdist)
            random_class_idx = random.randrange(0, len(possible_default_classes))
            default_class_att, default_class_value = list(classdist.keys())[random_class_idx]

            self.classifier = []
            self.default_class = default_class_value
            self.default_class_attribute = default_class_att
    
    def emptyrule(self):
        """returns rule with empty antecedent
        and consequent
        """
        return ClassAssocationRule(Antecedent([]), Consequent(None, None), 0, 0)
    
    
    def maxcoverrule(self, datacase, rules):
        """
        finds the highest precedence rule that covers
        the case d
        
        
        Arguments
        ---------
        rules: sorted rules
            
        datacase: instance d
            
        """
        crule, wrule = None, None
        
        
        for rule in rules:
            if rule.antecedent <= datacase:
                if rule.consequent == datacase.class_val and not crule:
                    # save cRule
                    crule = rule
                    if crule and wrule:
                        return crule, wrule
                elif rule.consequent != datacase.class_val and not wrule:
                    # save wRule
                    wrule = rule
                    if crule and wrule:
                        return crule, wrule

        
        
        return crule, wrule
    
    
    def allcover_rules(self, U, datacase, crule):
        """method for finding all rules from a set U
        that cover datacase and have a higher precedence
        tha crule
        """
        wset = set()
        
        for replacingrule in U:
            if replacingrule > crule and replacingrule.antecedent <= datacase and replacingrule.consequent.value != datacase.class_val.value:
                wset.add(replacingrule)
        
        return wset
    
    def errors_of_rule(self, rule):
        """method for computing errors of
        a rule
        """
        rule.support_count = sum(rule.class_cases_covered.values()) 
        return rule.support_count - rule.class_cases_covered[rule.consequent.value]
    
    
    
    def select_default_class(self, classdist):
        """method for selecting default class
        from class distribution
        """
        most_common = classdist.most_common(1)
        
        if not most_common:
            return (None, 0)
        
        return most_common[0]
    



from ..data_structures import TransactionDB
import matplotlib.pyplot as plt
import matplotlib.patches as patches

import pandas as pd
import numpy
import re



movies = pd.read_csv("../data/movies.csv", sep=";")
movies_discr = movies.copy(True)

budget_bins = range(0, 350, 50)
budget_bins_names = [ "<{0};{1})".format(i, i + 50) for i in budget_bins[:-1] ]

celebrities_bins = range(0, 10, 2)
celebrities_bins_names = [ "<{0};{1})".format(i, i + 2) for i in celebrities_bins[:-1] ]


transactionDB = TransactionDB.from_DataFrame(movies_discr, unique_transactions=True)

movies_vals = movies.get_values()

x = range(0, 350, 50)
y = range(1, 9)


x_points = list(map(lambda n: n[0], movies_vals))
y_points = list(map(lambda n: n[1], movies_vals))
data_class = list(movies['class'])
appearance = {
    'box-office-bomb': ('brown', "o"),
    'main-stream-hit': ('blue', "o"),
    'critical-success': ('green', "o")
}
rule_appearance = {
    'box-office-bomb': 'tan',
    'main-stream-hit': 'aqua',
    'critical-success': 'lightgreen'
}

plt.style.use('seaborn-white')


def plot_qrule(qrule, plt):
    interval_regex = "(?:<|\()(\d+(?:\.(?:\d)+)?);(\d+(?:\.(?:\d)+)?)(?:\)|>)"
    
    lower_y = 0
    area_y = celebrities_bins[-1]
    
    lower_x = 0
    area_x = budget_bins[-1]
    
    antecedent = qrule.antecedent
    
    
    if len(antecedent) != 0:
        if antecedent[0][0] == "a-list-celebrities":
            y = antecedent[0]
            y_boundaries = re.search(interval_regex, repr(y[1]))
            lower_y = float(y_boundaries.group(1))
            upper_y = float(y_boundaries.group(2))
            area_y = upper_y - lower_y

            axis = plt.gca()
        else:
            x = antecedent[0]
            x_boundaries = re.search(interval_regex, repr(x[1]))
            lower_x = float(x_boundaries.group(1))
            upper_x = float(x_boundaries.group(2))
            area_x = upper_x - lower_x
        
    if len(antecedent) > 1:
        if antecedent[1][0] == "a-list-celebrities":
            y = antecedent[0]
            y_boundaries = re.search(interval_regex, repr(y[1]))
            lower_y = float(y_boundaries.group(1))
            upper_y = float(y_boundaries.group(2))
            area_y = upper_y - lower_y

            axis = plt.gca()
        else:
            x = antecedent[1]
            x_boundaries = re.search(interval_regex, repr(x[1]))
            lower_x = float(x_boundaries.group(1))
            upper_x = float(x_boundaries.group(2))
            area_x = upper_x - lower_x
            
    
    axis = plt.gca()

    class_name = qrule.consequent[1]
    
    axis.add_patch(
       patches.Rectangle((lower_x, lower_y), area_x, area_y, zorder=-2, facecolor=rule_appearance[class_name], alpha=qrule.confidence)
    )







def plot_rule(rule, plt):
    interval_regex = "<(\d+);(\d+)\)"
    
    lower_y = 0
    area_y = celebrities_bins[-1]
    
    lower_x = 0
    area_x = budget_bins[-1]
    
    
    
    
    if len(rule.antecedent) != 0:
        if rule.antecedent[0][0] == "a-list-celebrities":
            y = rule.antecedent[0]
            y_boundaries = re.search(interval_regex, y[1])
            lower_y = float(y_boundaries.group(1))
            upper_y = float(y_boundaries.group(2))
            area_y = upper_y - lower_y

            axis = plt.gca()
        else:
            x = rule.antecedent[0]
            x_boundaries = re.search(interval_regex, x[1])
            lower_x = float(x_boundaries.group(1))
            upper_x = float(x_boundaries.group(2))
            area_x = upper_x - lower_x
        
    if len(rule.antecedent) > 1:
        if rule.antecedent[1][0] == "a-list-celebrities":
            y = rule.antecedent[1]
            y_boundaries = re.search(interval_regex, y[1])
            lower_y = float(y_boundaries.group(1))
            upper_y = float(y_boundaries.group(2))
            area_y = upper_y - lower_y

            axis = plt.gca()
        else:
            x = rule.antecedent[1]
            x_boundaries = re.search(interval_regex, x[1])
            lower_x = float(x_boundaries.group(1))
            upper_x = float(x_boundaries.group(2))
            area_x = upper_x - lower_x
            
    
    axis = plt.gca()

    class_name = rule.consequent[1]
    
    axis.add_patch(
       patches.Rectangle((lower_x, lower_y), area_x, area_y, zorder=-2, facecolor=rule_appearance[class_name], alpha=rule.confidence)
    )



def plot_quant_rules(qrules):
    for r in qrules:
        plot_qrule(r, plt)

    # data cases
    for i in range(len(x_points)):
        plt.scatter(x_points[i], y_points[i], marker=appearance[data_class[i]][1], color=appearance[data_class[i]][0], s=60)


    # rule boundary lines
    for i, n in enumerate(x):
        plt.axhline(y=y[i], color = "grey", linestyle="dashed")
        plt.axvline(x=x[i], color = "grey", linestyle="dashed")

        
        
    plt.xlabel('Estimated Budget (1000$)')
    plt.ylabel('A-List Celebrities')




def plot_rules(rules):
    for r in rules:
        plot_rule(r, plt)

    # data cases
    for i in range(len(x_points)):
        plt.scatter(x_points[i], y_points[i], marker=appearance[data_class[i]][1], color=appearance[data_class[i]][0], s=60)


    # rule boundary lines
    for i, n in enumerate(x):
        plt.axhline(y=y[i], color = "grey", linestyle="dashed")
        plt.axvline(x=x[i], color = "grey", linestyle="dashed")

        
        
    plt.xlabel('Estimated Budget (1000$)')
    plt.ylabel('A-List Celebrities')


class Appearance:
    """Appearance represents an easy way to get an appearance 
    dictionary for functions from fim package.


    Attributes
    ----------
    self.lhs: array of (string, string)

    self.rhs: array of (string, string)

    frozenset: frozenset of Items
        this attribute is vital for determining if antecedent
        is a subset of transaction and, consequently, if transaction
        satisfies antecedent

    """
    
    def __init__(self):
        self.lhs = []
        self.rhs = []
        
        
    def add_to_LHS(self, item):
        self.__add(item, "a")
        
    def add_to_RHS(self, item):
        self.__add(item, "c")
        
    def __add(self, item, where):
        """
        Function for adding a condition to either self.rhs
        or self.lhs
        """

        key, value = item.attribute, item.value
        string_repr = "{}:=:{}".format(key, value)
        
        # finding to which side we need to insert
        where_list = self.lhs if where == "a" else self.rhs
        
        # inserting a condition
        where_list.append((string_repr, where))
        
        
    @property
    def dictionary(self):
        """
        Get a final dictionary to be used in functions 
        from fim package.
        """
        if not self.lhs:
            self.lhs.append((None, "a"))
            
        if not self.rhs:
            self.rhs.append((None, "c"))
            
        appear_list = self.lhs + self.rhs
        
        return dict(appear_list)



import time
import fim
from ..data_structures import Consequent, Item, Antecedent, ClassAssocationRule

def createCARs(rules):
    """Function for converting output from fim.arules or fim.apriori
    to a list of ClassAssociationRules

    Parameters
    ----------
    rules : output from fim.arules or from generateCARs


    Returns
    -------
    list of CARs

    """
    CARs = []
    
    for rule in rules:
        con_tmp, ant_tmp, support, confidence = rule

        con = Consequent(*con_tmp.split(":=:"))

        # so that the order of items in antecedent is always the same
        ant_tmp = sorted(list(ant_tmp))
        ant_items = [ Item(*i.split(":=:")) for i in ant_tmp ]
        ant = Antecedent(ant_items)

        CAR = ClassAssocationRule(ant, con, support=support, confidence=confidence)
        CARs.append(CAR)

    CARs.sort(reverse=True)

    return CARs


def generateCARs(transactionDB, support=1, confidence=50, maxlen=10, **kwargs):
    """Function for generating ClassAssociationRules from a TransactionDB

    Parameters
    ----------
    transactionDB : TransactionDB

    support : float
        minimum support in percents if positive
        absolute minimum support if negative

    confidence : float
        minimum confidence in percents if positive
        absolute minimum confidence if negative

    maxlen : int
        maximum length of mined rules

    **kwargs : 
        arbitrary number of arguments that will be 
        provided to the fim.apriori function

    Returns
    -------
    list of CARs

    """
    appear = transactionDB.appeardict
    
    rules = fim.apriori(transactionDB.string_representation, supp=support, conf=confidence, mode="o", target="r", report="sc", appear=appear, **kwargs, zmax=maxlen)
    

    return createCARs(rules)
 

def top_rules(transactions,
              appearance={},
              target_rule_count=1000,
              init_support=0.,
              init_conf=0.5,
              conf_step=0.05,
              supp_step=0.05,
              minlen=2,
              init_maxlen=3,
              total_timeout=100.,
              max_iterations=30):
    """Function for finding the best n (target_rule_count)
    rules from transaction list

    Parameters
    ----------
    transactions : 2D array of strings
        e.g. [["a:=:1", "b:=:3"], ["a:=:4", "b:=:2"]]

    appearance : dictionary
        dictionary specifying rule appearance

    targent_rule_count : int
        target number of rules to mine

    init_conf : float
        confidence from which to start mining

    conf_step : float

    supp_step : float

    minen : int
        minimum len of rules to mine

    init_maxlen : int
        maxlen from which to start mining

    total_timeout : float
        maximum execution time of the function

    max_iterations : int
        maximum iterations to try before stopping
        execution


    Returns
    -------
    list of mined rules. The rules are not ordered.

    """
    
    starttime = time.time()
    
    MAX_RULE_LEN = len(transactions[0])
    
    support = init_support
    conf = init_conf
    
    maxlen = init_maxlen
    
    flag = True
    lastrulecount = -1
    maxlendecreased_due_timeout = False
    iterations = 0
    
    rules = None
    
    
    while flag:
        iterations += 1
            
        if iterations == max_iterations:
            print("Max iterations reached")
            break
                
                
        
        print("Running apriori with setting: confidence={}, support={}, minlen={}, maxlen={}, MAX_RULE_LEN={}".format(
                conf, support, minlen, maxlen, MAX_RULE_LEN))
        
        rules_current = fim.arules(transactions, supp=support, conf=conf, mode="o", report="sc", appear=appearance, zmax=maxlen, zmin=minlen)
        
        rules = rules_current
        
        rule_count = len(rules)
        
        print("Rule count: {}, Iteration: {}".format(rule_count, iterations))
        
        if (rule_count >= target_rule_count):
            flag = False
            print("Target rule count satisfied:", target_rule_count)
        else:
            exectime = time.time() - starttime
            
            if exectime > total_timeout:
                print("Execution time exceeded:", total_timeout)
                flag = False
                
            elif maxlen < MAX_RULE_LEN and lastrulecount != rule_count and not maxlendecreased_due_timeout:
                    maxlen += 1
                    lastrulecount = rule_count
                    print("Increasing maxlen", maxlen)
                        
            elif maxlen < MAX_RULE_LEN and maxlendecreased_due_timeout and support <= 1 - supp_step:
                support += supp_step
                maxlen += 1
                lastrulecount = rule_count
                
                print("Increasing maxlen to", maxlen)
                print("Increasing minsup to", support)
                
                maxlendecreased_due_timeout = False
            
            elif conf > conf_step:
                conf -= conf_step
                print("Decreasing confidence to", conf)
                
            else:
                print("All options exhausted")
                flag = False
        
            
       
    return rules






from functools import reduce
from .comparable_itemset import ComparableItemSet

class Antecedent(ComparableItemSet):
    """Antecedent represents a left-hand side of the association rule.
    It is a set of conditions (Items) a Transaction has to satisfy.

    Parameters
    ----------
    items: 1D array of Items


    Attributes
    ----------
    itemset: 1D array of Items
        dictionary of unique attributes, such as: {a: 1, b: 3}

    frozenset: frozenset of Items
        this attribute is vital for determining if antecedent
        is a subset of transaction and, consequently, if transaction
        satisfies antecedent

    """
    
    def __init__(self, items):

        # extract unique attributes and convert them to dict
        # such as: {a: 1, b: 3, c: 4}
        self.itemset = dict(list(set(items)))

        # this part is important for better performance
        # of M1 and M2 algoritms
        self.frozenset = frozenset(self)
        
    
    def __getattr__(self, attr_name):
        """
        Parameters
        ----------
        attribute: str
            name of desired attribute

        Returns
        -------
        Attribute of given name, otherwise an AttributeError
        """
        item = self.itemset.get(attr_name, None)
        
        if (item):
            return item
        else:
            raise AttributeError("No attribute of that name")
            
    
    def __getitem__(self, idx):
        """Method which allows indexing on antecedent's itemset
        """
        items = list(self.itemset.items())
        
        if (idx <= len(items)):
            return items[idx]
        else:
            raise IndexError("No value at the specified index")

    def __len__(self):
        """
        Returns
        -------
        length of the itemset
        """
        return len(self.itemset)
            
    def __repr__(self):
        str_array = [repr((attr, val)) for attr, val in self.itemset.items()]
        text = ", ".join(str_array)
        return "Antecedent({})".format(text)
    
    def __hash__(self):
        return hash(tuple(self.itemset.items()))
    
    def __eq__(self, other):
        return hash(self) == hash(other)

    def string(self):
        items = list(self.itemset.items())
        string_items = [ "{}={}".format(key, val) for key, val in items ]

        string_ant = ",".join(string_items)

        return "{" + string_ant + "}"


from .algorithms import (
    M1Algorithm,
    M2Algorithm,
    generateCARs,
    createCARs,
    top_rules
)
from .data_structures import TransactionDB

class CBA():
    """Class for training a testing the
    CBA Algorithm.

    Parameters:
    -----------
    support : float
    confidence : float
    algorithm : string
        Algorithm for building a classifier.
    maxlen : int
        maximum length of mined rules
    """


    def __init__(self, support=0.10, confidence=0.5, maxlen=10, algorithm="m1"):
        if algorithm not in ["m1", "m2"]:
            raise Exception("algorithm parameter must be either 'm1' or 'm2'")
        if 0 > support or support > 1:
            raise Exception("support must be on the interval <0;1>")
        if 0 > confidence or confidence > 1:
            raise Exception("confidence must be on the interval <0;1>")
        if maxlen < 1:
            raise Exception("maxlen cannot be negative or 0")
        
        self.support = support * 100
        self.confidence = confidence * 100
        self.algorithm = algorithm
        self.maxlen = maxlen
        self.clf = None
        
        self.available_algorithms = {
            "m1": M1Algorithm,
            "m2": M2Algorithm
        }

        
        


    def rule_model_accuracy(self, txns):
        """Takes a TransactionDB and outputs
        accuracy of the classifier
        """
        if not self.clf:
            raise Exception("CBA must be trained using fit method first")
        if not isinstance(txns, TransactionDB):
            raise Exception("txns must be of type TransactionDB")

        return self.clf.test_transactions(txns)
        
    def fit(self, transactions, top_rules_args={}):
        """Trains the model based on input transaction
        and returns self.
        """
        if not isinstance(transactions, TransactionDB):
            raise Exception("transactions must be of type TransactionDB")

        used_algorithm = self.available_algorithms[self.algorithm]
        
        cars = None

        if not top_rules_args:
            cars = generateCARs(transactions, support=self.support, confidence=self.confidence, maxlen=self.maxlen)
        else:
            rules = top_rules(transactions.string_representation, appearance=transactions.appeardict, **top_rules_args) 
            cars = createCARs(rules)

        self.clf = used_algorithm(cars, transactions).build()
        
        return self
    
    def predict(self, X):
        """Method that can be used for predicting
        classes of unseen cases.

        CBA.fit must be used before predicting.
        """
        if not self.clf:
            raise Exception("CBA must be train using fit method first")

        if not isinstance(X, TransactionDB):
            raise Exception("X must be of type TransactionDB")


        return self.clf.predict_all(X)


    def predict_probability(self, X):
        """Method for predicting probablity of 
        given classification
¨
        CBA.fit must be used before predicting probablity.
        """

        return self.clf.predict_probability_all(X)

    def predict_matched_rules(self, X):
        """for each data instance, returns a rule that
        matched it according to the CBA order (sorted by 
        confidence, support and length)
        """

        return self.clf.predict_matched_rule_all(X)
    
    
    




from .appearance import Appearance
from .transaction import Transaction, UniqueTransaction
from . import Item

class TransactionDB:
    
    def __init__(self, dataset, header, unique_transactions=True, drop_NaN=True):
        """TransactionDB represents a list of Transactions that can be
        passed to CBA algorithm as a training or a test set. 

        Parameters
        ----------
        
        dataset: two dimensional array of strings or ints
    
        header: array of strings
            Represents column labels.
        
        unique_transactions: bool
            Determines if UniqueTransaction or Transaction class
            should be used for individual instances.

        drop_NaN: bool
            Used for determining whether a an Item
            with NULL value should be dropped from Transaction


        Attributes
        ----------
        header: array of strings
            Column labels.

        class_labels: array of Items

        classes: array of strings
            Individual values of class_labels.

        data: array of Transactions
            Individual instances.

        string_representation: two dimensional array of strings
            e.g. [["food:=:schitzel", "mood:=:happy"], ["food:=:not_schitzel], ["mood:=:unhappy"]]


        """
        
        TransactionClass = UniqueTransaction if unique_transactions else Transaction
        
        self._dataset_param = dataset
        self.header = header
        self.class_labels = []
        
        new_dataset = []

        for row in dataset:
            class_label = Item(header[-1], row[-1])
            new_row = TransactionClass(row[:-1], header[:-1], class_label, drop_NaN=drop_NaN)
            
            self.class_labels.append(class_label)
            
            new_dataset.append(new_row)
            
        self.data = new_dataset
        self.classes = list(map(lambda i: i[1], self.class_labels))
        
        
        
        get_string_items = lambda transaction: transaction.string_items
        
        mapped = map(get_string_items, self)
        
        self.string_representation = list(mapped)
        
        

    @property
    def appeardict(self):
        """
        Returns
        -------
        an appearance dictionary to be used in the fim
        package. Assumes user wants to generate class association
        rules.
        """
        appear = Appearance()
        
        unique_class_items = set(self.class_labels)
        
        for item in unique_class_items:
            appear.add_to_RHS(item)

        return appear.dictionary


    @property
    def appeardict_itemsets_only(self):
        """
        Returns
        -------
        an appearance dictionary to be used in the fim
        package. Assumes user wants to generate frequent itemsets
        only, not class assocation rules
        """
        appear = Appearance()
        
        return appear.dictionary
        
    
    def __getitem__(self, idx):
        return self.data[idx]
    
    
    @classmethod
    def from_DataFrame(clazz, df, unique_transactions=False, drop_NaN=True, target=None):
        """
        Allows the conversion of pandas DataFrame class to 
        TransactionDB class.

        Parameters
        ----------
        
        df: pandas DataFrame
            A DataFrame from which to create a TransactionDB.
    
        unique_transactions: bool
            Determines if UniqueTransaction or Transaction class
            should be used for individual instances.

        drop_NaN: bool
            Used for determining whether a an Item
            with NULL value should be dropped from Transaction.

        target: str, default None
            Name of an existing column in df. This column will
            be taken as a class target.
        """

        if target is not None:
            if type(target) != str:
                raise Exception("'target' should be a string")

            if target not in df.columns.values:
                raise Exception("'target' must be in df columns")

            new_columns = list(df.columns.values)
            new_columns.pop(new_columns.index(target))
            new_columns.append(target)

            df = df[new_columns]
        
        rows = df.values
        header = list(df.columns.values)

        return clazz(rows, header, unique_transactions=unique_transactions, drop_NaN=drop_NaN)

    
    def __repr__(self):
        return repr(self.string_representation)
        
    def __len__(self):
        return len(self.data)
        



class ComparableItemSet:
    """ ComparableItemSet is a common ancestor
    for Antecedent and Transaction class so that
    they both can be compared using <= and >= 
    operators.

    Any class that inherits from ComparableItemSet
    needs to have a "frozenset" attribute. "frozenset"
    attribute is a frozenset of Items and allows easy 
    comparing and determining if one ComparableItemSet
    is a subset or superset of another ComparableItemSet.

    """

    def issuperset(self, other):
        return self.frozenset >= other.frozenset
        
    def issubset(self, other):
        return self.frozenset <= other.frozenset 
        
    def __ge__(self, other):
        return self.issuperset(other)
        
    def __le__(self, other):
        return self.issubset(other)
    
    


from .data_structures import QuantitativeDataFrame
from sklearn.metrics import accuracy_score

class QuantitativeClassifier:
    
    def __init__(self, rules, default_class):
        self.rules = rules
        self.default_class = default_class
        
        
    def rule_model_accuracy(self, quantitative_dataframe, ground_truth):
        predicted = self.predict(quantitative_dataframe)

        return accuracy_score(predicted, ground_truth)

    def predict(self, quantitative_dataframe):
        predicted_classes = []
    
        for _, row in quantitative_dataframe.dataframe.iterrows():
            appended = False
            for rule in self.rules:
                antecedent_dict = dict(rule.antecedent)  
                counter = True

                for name, value in row.iteritems():
                    if name in antecedent_dict:
                        interval = antecedent_dict[name]

                        if type(interval) == str:
                            counter &= interval == value
                        else:
                            result = interval.isin(value)
                            counter &= result

                if counter:
                    _, predicted_class = rule.consequent
                    predicted_classes.append(predicted_class)
                    appended = True
                    break
                    
            if not appended:
                predicted_classes.append(self.default_class)

                    
        return predicted_classes            






from .comparable_itemset import ComparableItemSet
from .item import Item
import pandas as pd
import numpy as np


class Transaction(ComparableItemSet):
    """Transaction represents one instance in a dataset.
    Transaction is hashed based on its items and class 
    value. 

    Parameters
    ----------
    
    row: array of ints or strings

    header: array of strings
        Represents column labels.
    
    class_item: Item
        Item with class attribute.

    drop_NaN: bool
        Used for determining whether a an Item
        with NULL value should be dropped from Transaction


    Attributes
    ----------
    items: array of Items

    tid: int
        Transaction ID.

    alreadycovered: bool
        Used in M2Algorithm for determining if the transaction
        was already covered by some other rule.

    string_items: two dimensional array of strings
        e.g. [["a:=:1", "b:=:2"]]




    """


    id_ = 0
    
    def __init__(self, row, header, class_item, drop_NaN=True):
        self.class_val = class_item
        self.items = []
        self.tid = Transaction.id_
        Transaction.id_ += 1
        
        self.alreadycovered = False
        
        # eg. [pay=high, eyes=green]
        self.string_items = []
        
        
        for idx, val in enumerate(row):
            # Drop items with NULL value
            if drop_NaN and pd.isnull(val):
                continue

            header_label = header[idx]
            
            item = Item(header_label, val)
            
            self.string_items.append("{}:=:{}".format(header_label, val)) 
            
            self.items.append(item)
            
        key, val = self.class_val
        self.string_items.append("{}:=:{}".format(key, val))

        self.frozenset = frozenset(self)
            
            
    
    def __repr__(self):
        string = ", ".join(self.string_items) 
        return "{" + string + "}"
    
    def __hash__(self):
        return hash(tuple(self.items))
        #return hash((self.tid, tuple(self.items)))

    def __eq__(self, other):
        return hash(self) == hash(other)
    
    def __getitem__(self, idx):
        return self.items[idx]
    
    def getclass(self):
        return self.class_val



class UniqueTransaction(Transaction):
    """Same as Transaction class except for
    hashing by Transaction id. 

    """
    def __hash__(self):
        return hash(self.tid)


from .transforms import (
    RuleRefitter,
    RuleLiteralPruner,
    RuleTrimmer,
    RuleExtender,
    RulePostPruner,
    RuleOverlapPruner
)


class QCBATransformation:


    def __init__(self, quantitative_dataset, transaction_based_drop=True):
        self.transaction_based_drop = transaction_based_drop

        self.dataset = quantitative_dataset

        self.refitter = RuleRefitter(self.dataset)
        self.literal_pruner = RuleLiteralPruner(self.dataset)
        self.trimmer = RuleTrimmer(self.dataset)
        self.extender = RuleExtender(self.dataset)
        self.post_pruner = RulePostPruner(self.dataset)
        self.overlap_pruner = RuleOverlapPruner(self.dataset)


    def transform(self, rules, transformation_dict={}):

        if not transformation_dict:
            print("applying all transformations")
            refitted = self.refitter.transform(rules)
            literal_pruned = self.literal_pruner.transform(refitted)
            trimmed = self.trimmer.transform(literal_pruned)
            extended = self.extender.transform(trimmed)
            post_pruned, default_class = self.post_pruner.transform(extended)
            overlap_pruned = self.overlap_pruner.transform(post_pruned, default_class, transaction_based=self.transaction_based_drop)
        
        else:
            print("applying selected transformations")
            transformed_rules = rules
 
            if transformation_dict.get("refitting", False):
                print("refitting")
                transformed_rules = self.refitter.transform(transformed_rules)
            if transformation_dict.get("literal_pruning", False):
                print("literal pruning")
                transformed_rules = self.literal_pruner.transform(transformed_rules)
            if transformation_dict.get("trimming", False):
                print("trimming")
                transformed_rules = self.trimmer.transform(transformed_rules)
            if transformation_dict.get("extension", False):
                print("extending")
                transformed_rules = self.extender.transform(transformed_rules)

            print("post pruning")
            transformed_rules, default_class = self.post_pruner.transform(transformed_rules)
            
            if transformation_dict.get("overlap_pruning", False):
                print("overlap pruning")
                transaction_based = transformation_dict["transaction_based_drop"]

                transformed_rules = self.overlap_pruner.transform(transformed_rules, default_class, transaction_based=transaction_based)

            return transformed_rules, default_class


        return overlap_pruned, default_class




from collections import Counter

class RuleBuilderAlgorithm:
    """Common ancestor for M1 and M2 Algorithms
    to provide common interface.
    """

    def __init__(self, rules, dataset):
        self.rules = rules
        self.dataset = dataset
        self.y = dataset.class_labels
        
    def update_class_distr(self, classdist, rule):
        return classdist - rule.class_cases_covered

    def calculate_default_class_properties(self, clf):
        """This function is used for calculating
        default class support and confidence
        """
        default_class = clf.default_class
        class_distribution = Counter([ value for _, value in self.y])

        clf.default_class_support = class_distribution[default_class] / len(self.y)
        clf.default_class_confidence = class_distribution[default_class] / len(self.y)

    


import pandas as pd
from functools import reduce

from ..data_structures import ClassAssocationRule, Antecedent, Consequent

class Classifier:
    """
    Classifier for CBA that can predict 
    class labels based on a list of rules.
    """

    def __init__(self):
        self.rules = []
        self.default_class = None
        self.default_class_attribute = None
        self.default_class_confidence = None
        self.default_class_support = None



    def test_transactions(self, txns):
        """Takes a TransactionDB and outputs
        accuracy of the classifier
        """
        pred = self.predict_all(txns)
        actual = txns.classes

        return accuracy_score(pred, actual)
    
        
    def predict(self, datacase):
        """predicts target class of one 
        datacase
        """
        for rule in self.rules:
            if rule.antecedent <= datacase:
                return rule.consequent.value
            
        return self.default_class
        
    def predict_all(self, dataset):
        """predicts target class of an array
        of datacases
        """
        predicted = []
        
        for datacase in dataset:
            predicted.append(self.predict(datacase))
            
        return predicted

    def predict_matched_rule(self, datacase):
        """returns a rule that matched the instance
        according to the CBA order (rules are sorted
        by confidence, support and length and first matched
        rule is returned)
        """
        for rule in self.rules:
            if rule.antecedent <= datacase:
                return rule

        default_rule_ant = Antecedent({})
        default_rule_conseq = Consequent(self.default_class_attribute, self.default_class)
        
        default_rule = ClassAssocationRule(
            default_rule_ant,
            default_rule_conseq,
            self.default_class_support,
            self.default_class_confidence
        )

        return default_rule

    def predict_matched_rule_all(self, dataset):
        """for each data instance, returns a rule that
        matched it according to the CBA order (sorted by 
        confidence, support and length)
        """
        matched_rules = []
        
        for datacase in dataset:
            matched_rules.append(self.predict_matched_rule(datacase))
            
        return matched_rules



    def predict_probability(self, datacase):
        """predicts target class probablity of one 
        datacase
        """
        for rule in self.rules:
            if rule.antecedent <= datacase:
                return rule.confidence
            
        return self.default_class_confidence

    def predict_probability_all(self, dataset):
        """predicts target class probablity
        of an array of datacases
        """
        predicted = []
        
        for datacase in dataset:
            predicted.append(self.predict_probability(datacase))
            
        return predicted


    def inspect(self):
        """inspect uses pandas DataFrame to
        display information about the classifier
        """
        
        dictionary = {
            "lhs": [],
            "rhs": [],
            "confidence": [],
            "support": [],
            "length": [],
            "id": []
        }

        for rule in self.rules:
            dictionary["lhs"].append(rule.antecedent.string())
            dictionary["rhs"].append(rule.consequent.string())
            dictionary["confidence"].append(rule.confidence)
            dictionary["support"].append(rule.support)
            dictionary["length"].append(len(rule.antecedent) + 1)
            dictionary["id"].append(rule.rid)

        # default rule
        dictionary["lhs"].append("{}")
        dictionary["rhs"].append(self.default_class)
        dictionary["confidence"].append(self.default_class_confidence)
        dictionary["support"].append(self.default_class_support)
        dictionary["length"].append(1)
        dictionary["id"].append(None)


        rules_df = pd.DataFrame(dictionary)
        rules_df = rules_df[["lhs", "rhs", "confidence", "support", "length", "id"]]

        return rules_df



def accuracy_score(actual, predicted):
    """Function for determining accuracy given
    list of predicted classes and actual classes
    """

    length = len(actual)

    indices = range(length)

    def reduce_indices(previous, current):
        i = current

        result = 1 if actual[i] == predicted[i] else 0

        return previous + result

    accuracy = reduce(reduce_indices, indices) / length

    return accuracy


import collections
from .rule_algorithm import RuleBuilderAlgorithm
from .classifier import Classifier

import time
import random

class M1Algorithm(RuleBuilderAlgorithm):
    """ M1 Algorithm implementation.
    """
    
    def build(self):
        
        # list for storing rules to be used in the classifier
        classifier = []
        # list for storing default classes associated
        # with rules in the classifier
        default_classes = []
        # list for storing errors of said default classes
        default_classes_errors = []
        # list for storing rule errors from classifier
        rule_errors = []
        # list for storing total errors
        # (rule_errors + default_classes_errors)
        total_errors = []
        # class distribution
        # for calculating the default's rule confidence
        # and support
        class_distribution = collections.Counter(self.y)
        classdist_keys = list(class_distribution.keys())


        # sorting rules based on the precedence operator
        self.rules.sort(reverse=True)

        # converting TransactionDB to a set
        # so that set intersection and difference can be used
        dataset = set(self.dataset)

        # obtaining the set's length. We do this only once to
        # save processing time.
        # this is a constant variable
        dataset_len = len(dataset)

        # When we want to update the dataset_len, we use
        # this variable. Length is updated by subtracting 
        # absolute support of a rule from it
        dataset_len_updated = dataset_len
        
        

        for rule in self.rules:
            # if all data cases have been covered
            # break the loop to save time
            if (dataset_len_updated <= 0):
                break
            

            # temp serves for storing datacases
            # that have been covered by current rule
            temp = set()
            # temp len is for determining temp's length
            # without using len(temp) to save time
            temp_len = 0
            # number of rule that satisfy both antecedent
            # and consequent of the current rule
            temp_satisfies_conseq_cnt = 0
            
            
            for datacase in dataset:
                # if datacase satisfies rule's antecedent
                # we'll store it in temp and increment
                #  temp's len
                if rule.antecedent <= datacase:
                    temp.add(datacase)
                    temp_len += 1

                    # we'll mark the rule if datacase
                    # satisfies its consequent. And increment
                    # the counter
                    if rule.consequent == datacase.class_val:  
                        temp_satisfies_conseq_cnt += 1
                        rule.marked = True
                        

            # if rule satisfied at least one consequent
            if rule.marked:
                classifier.append(rule)

                # we subtract already covered rules
                # from dataset                
                dataset -= temp
                # and update dataset's length
                dataset_len_updated -= temp_len
                
                # we'll obtain Counter of remaining class values
                # in the dataset using map to save time
                class_distribution = collections.Counter(map(lambda d: d.class_val.value, dataset))
                
                # the most common value from the counter will be
                # the default class
                most_common_tuple = class_distribution.most_common(1)
                

                # here we'll do some checking in case
                # the counter is empty
                most_common_cnt = 0
                most_common_label = "None"
                
                try:
                    most_common_tuple = most_common_tuple[0]
                    most_common_cnt = most_common_tuple[1]
                    most_common_label = most_common_tuple[0]
                except IndexError:
                    pass
                
                    
                # the most common label will be inserted into 
                # the list                
                default_classes.append(most_common_label)
                
                
                # number of errors the rule will make => 
                #
                # difference of:
                # all transactions that satisfy its antecedent
                # and
                # all transactions that satisfy both antecedent and consequent
                rule_errors.append(temp_len - temp_satisfies_conseq_cnt)
                
                # default errors
                #
                # difference of:
                # length of remaining dataset
                # and
                # count of most common class 
                dflt_class_err = dataset_len_updated - most_common_cnt
                
                
                err_cnt = dflt_class_err
                    
                
                default_classes_errors.append(err_cnt)
                
                total_errors.append(err_cnt + sum(rule_errors))
                
                
        
        # finding the smallest number of errors
        # but checking if at least one rule classified an instance
        if len(total_errors) != 0:            
            min_errors = min(total_errors)
            
            # finding the index of smallest number of errors
            idx_to_cut = total_errors.index(min_errors)
            
            final_classifier = classifier[:idx_to_cut+1]
            default_class = default_classes[idx_to_cut]        
            
            # creating the final classifier
            clf = Classifier()
            clf.rules = final_classifier
            clf.default_class = default_class
            clf.default_class_attribute = classdist_keys[0][0]

        else:
            clf = Classifier()
            clf.rules = []

            possible_default_classes = list(class_distribution)
            random_class_idx = random.randrange(0, len(possible_default_classes))
            default_class_att, default_class_value = classdist_keys[random_class_idx]
            clf.default_class = default_class_value
            clf.default_class_attribute = default_class_att


        self.calculate_default_class_properties(clf)        

        return clf


        


import collections

class ClassAssocationRule():
    """ClassAssociationRule (CAR) is defined by its antecedent, consequent,
    support, confidence and id. 

    It has a set of Items in its antecedent and one Item in its
    Consequent. 

    __lt__ and __gt__ operators are overriden so that list of CARs can
    be sorted.  
    
    
    Parameters
    ----------
    
    antecedent: Antecedent
        Items that a Transaction has to satisfy
   
    consequent: Consequent
        Target class of a Transaction that satisfies
        antecedent
    
    support: float
        how many transactions satisfy the rule, relatively
    
    confidence: float
        relative degree of certainty that consequent holds
        given antecedent



    Attributes
    ----------
    antecedent

    conseqent

    support

    confidence

    rid: int
        rule id

    support_count: int
        absolute support count

    marked: bool

    class_cases_covered: collections.Counter
        counter for determining which transactions are
        covered by the antecedent. Important for M2Algorithm.
    
    replace: set of ClassAssociationRule
        set of rules that have higher precedence than
        this rule and can replace it in M2Algorithm.


    """

    id = 0

    def __init__(self, antecedent, consequent, support, confidence):
        self.antecedent = antecedent
        self.consequent = consequent
        self.support = support
        self.confidence = confidence
        self.rulelen = len(antecedent) + 1
        self.rid = ClassAssocationRule.id

        ClassAssocationRule.id += 1

        self.support_count = 0
        
        self.marked = False
        
        self.class_cases_covered = collections.Counter()
        self.replace = set()
        
        
    def __gt__(self, other):
        """
        precedence operator. Determines if this rule
        has higher precedence. Rules are sorted according
        to their confidence, support, length and id.
        """
        if (self.confidence > other.confidence):
            return True
        elif (self.confidence == other.confidence and
              self.support > other.support):
            return True
        elif (self.confidence == other.confidence and
              self.support == other.support and
              self.rulelen < other.rulelen):
            return True
        elif(self.confidence == other.confidence and
              self.support == other.support and
              self.rulelen == other.rulelen and
              self.rid < other.rid):
            return True
        else:
            return False
    
    def __lt__(self, other):
        """
        rule precedence operator
        """
        return not self > other
    
    def __len__(self):
        """
        returns
        -------
        
        length of this rule 
        """
        return len(self.antecedent) + len(self.consequent)


    def __repr__(self):
        args = [self.antecedent.string(), "{" + self.consequent.string() + "}", self.support, self.confidence, self.rulelen, self.rid]
        text = "CAR {} => {} sup: {:.2f} conf: {:.2f} len: {}, id: {}".format(*args)

        return text


from .data_structures import QuantitativeCAR
from .transformation import QCBATransformation
from .classifier import QuantitativeClassifier

class QCBA:

    def __init__(self, quantitative_dataset, cba_rule_model=None, rules=None):
        if rules and cba_rule_model:
            raise Exception("rules and cba_rule_model cannot be specified together")

        if not rules and not cba_rule_model:
            raise Exception("either rules and cba_rule_model need to be specified")

        self.quantitative_dataset = quantitative_dataset
        self.__quant_rules = None

        if cba_rule_model:
            self.__quant_rules = [ QuantitativeCAR(r) for r in cba_rule_model.clf.rules ]
        if rules:
            self.__quant_rules = [ QuantitativeCAR(r) for r in rules ]
         

        self.qcba_transformation = QCBATransformation(quantitative_dataset)

        self.clf = None

    def fit(
            self, 
            refitting=True,
            literal_pruning=True,
            trimming=True,
            extension=True,
            overlap_pruning=True,
            transaction_based_drop=True
        ):

        transformation_dict = {
            "refitting": refitting,
            "literal_pruning": literal_pruning,
            "trimming": trimming,
            "extension": extension,
            "overlap_pruning": overlap_pruning,
            "transaction_based_drop": transaction_based_drop
        }

        transformed_rules, default_class = self.qcba_transformation.transform(self.__quant_rules, transformation_dict)

        self.clf = QuantitativeClassifier(transformed_rules, default_class)

        return self.clf


    def score(self, quantitative_dataset):
        actual = quantitative_dataset.dataframe.iloc[:, -1]
        
        return self.clf.rule_model_accuracy(quantitative_dataset, actual)


from .comparable_itemset import ComparableItemSet
from .item import Item

class Consequent(Item, ComparableItemSet):
    """
    Represents a right-hand side of the association rule.
    """
    
    def getclass(self):
        return self.value
    
    def __len__(self):
        return 1
    
    def __repr__(self):
        item_tuple = self.attribute, self.value
        return "Consequent{{{}}}".format(item_tuple)


class Item():
    """ Item class for representing attribute-value pair
    and one item in transaction or antecedent.


    Parameters
    ----------
    attribute : str
        name of the item

    value: str
        value of the item


    Attributes
    ----------
    attribute : str
        name of the item

    value: str
        value of the item


    """
    def __init__(self, attribute, value):
        # convert attribute and value so that 
        # Item("a", 1) == Item("a", "1")
        self.attribute = repr(attribute) if type(attribute) != str else attribute
        self.value = repr(value) if type(value) != str else value
        
    def __get_tuple(self):
        """Private method for getting an (attribute, value) pair"""
        return (self.attribute, self.value)
    
    def __getitem__(self, idx):
        """Method for accessing Item as a tuple"""
        item = self.__get_tuple()
        return item[idx]
    
    
    def __hash__(self):
        """Two Items with the same attribute and value
        have identical hash value.
        """
        return hash(self.__get_tuple())
    
    def __eq__(self, other):
        """Overriden method in order to compare based on
        value and not reference.
        """
        return hash(self) == hash(other)
    
    def __repr__(self):
        """Method for representing Item as a string.

        >>> item1 = Item("a", 1)
        >>> repr(item1)
        >>> Item{(a, 1)}
        """

        return "Item{{{}}}".format(self.__get_tuple())

    def string(self):
        """Method for getting simpler representation.
        
        
        >>> item1 = Item("a", 1)
        >>> item1.string()
        >>> a=1
        """
        return "{}={}".format(*self)
    
    


import sys
import importlib.util

package_name = 'fim'

spec = importlib.util.find_spec(package_name)
if spec is None:
    raise Exception("Before using pyARC, the 'fim' package must be installed first. For installation guide, please refer to http://www.borgelt.net/pyfim.html.")

from .cba import CBA
from .data_structures import TransactionDB